// cswitch_inout.c

#include "cswitch_inout.h"
#include "all_vtimers.h"
#include "vtimer.h"
#include "stacklist.h"
#include "hash_table.h"
#include "hwcounter_p4.h"

//#include <linux/kernel.h> // printk
#include <asm/current.h>
#include <linux/sched.h> // struct task_struct
#include <linux/spinlock.h>
#include <linux/stddef.h> // NULL
#include <linux/smp.h> // smp_processor_id()

extern unsigned give_debug_output;
extern struct all_vtimers av_vtimers;
extern struct hash_table ht_vtimers;
extern struct stacklist sl_heap;
extern spinlock_t kerninst_cswitch_lock;
extern unsigned cswitch_stuff_enabled;

// -------------------- Generic vstop/vrestart Routines ---------------------
void generic_vstop_routine(struct vtimer *vtimer, 
                           struct saved_timer *stimer,
                           unsigned type)
{
   struct vtimer_start *vt_start;
   unsigned counter_low=0, counter_high=0;
   uint64_t counter = 0;
   unsigned ctr_addr = 0;

   // Read appropriate hw_counter based on type
   if(type == 1) { /* HWC_TICKS */ 
      rdtsc(counter_low, counter_high);
   }
#ifdef NOT_FOR_DEMO
   else {
      ctr_addr = COUNTER_ADDR(type >> 16);
      __asm__ __volatile__ ("rdmsr"
			    : "=a" (counter_low), "=d" (counter_high)
			    : "c" (ctr_addr));
   }
#endif // NOT_FOR_DEMO
   counter = (uint64_t)(counter_high & 0x0000FFFFUL);
   counter <<= 32;
   counter |= (uint64_t)counter_low;

   // Store previous rcount and stop time in stimer.auxdata
   vt_start = (struct vtimer_start*)&stimer->auxdata;
   vt_start->start = counter;
   vt_start->rcount = vtimer->start_field.rcount;

   // Update vtimer total and zero out start_field
   vtimer->total_field += (counter - vtimer->start_field.start);
   vtimer->start_field.start = 0;
   vtimer->start_field.rcount = 0;
}

void generic_vrestart_routine(struct vtimer *vtimer, uint64_t auxdata,
                              unsigned type, unsigned wall)
{
   struct vtimer_start *old_start;
   unsigned counter_low=0, counter_high=0;
   uint64_t counter = 0;
   unsigned ctr_addr = 0;

   // Read appropriate hw_counter based on type
   if(type == 1) { /* HWC_TICKS */ 
      rdtsc(counter_low, counter_high);
   }
#ifdef NOT_FOR_DEMO
   else {
      ctr_addr = COUNTER_ADDR(type >> 16);
      __asm__ __volatile__ ("rdmsr"
			    : "=a" (counter_low), "=d" (counter_high)
			    : "c" (ctr_addr));
   }
#endif // NOT_FOR_DEMO
   counter = (uint64_t)(counter_high & 0x0000FFFFUL);
   counter <<= 32;
   counter |= (uint64_t)counter_low;

   vtimer->start_field.start = counter;
   old_start = (struct vtimer_start*)&auxdata;
   vtimer->start_field.rcount = old_start->rcount;

   // If wall timer, add events switched out to total
   if(wall) {
      vtimer->total_field += (counter - old_start->start);
   }   
}

// -------------------- Context Switch In/Out Routines ---------------------

// The following define is based on knowledge of how vtimers are allocated
// by the global data heap of kperfmon. If that allocation method ever changes,
// we will need to make sure we're still able to get the right vtimer for
// the given cpu.
#define vtimer_primitive_stride_in_bytes 64 

struct vtimer* getCPUTimer(struct vtimer *vt_base, unsigned cpu) {
   char *tmp = (char *)vt_base;
   tmp += (cpu * vtimer_primitive_stride_in_bytes);
   return (struct vtimer*)tmp;
}

void on_switch_in(void)
{
   struct vtimer *theTimer;
   struct saved_timer *timers_iter, *timers_start;
   uint64_t auxdata;
   struct task_struct *curr = current;
   unsigned curr_thr = (unsigned)curr->pid;
   unsigned curr_cpu = (unsigned)smp_processor_id();

   if(cswitch_stuff_enabled) {
      spin_lock(&kerninst_cswitch_lock);
   
      // From the hash table, get the stored vector-of-timer-ptrs for current
      // thread. If none exists, then we don't need to do anything more

      timers_start = hash_table_remove(&ht_vtimers, curr_thr);

      if (timers_start == NULL) { // probably the common case
         spin_unlock(&kerninst_cswitch_lock);
         return;
      }

      timers_iter = timers_start;
      theTimer = (struct vtimer *)(timers_iter->addr);
      auxdata = timers_iter->auxdata;

      while(theTimer != NULL) {
         // call the restart routine
         generic_vrestart_routine(getCPUTimer(theTimer, curr_cpu), auxdata,
                                  theTimer->counter_type, theTimer->wall_timer);

         timers_iter++;
         theTimer = (struct vtimer *)(timers_iter->addr);
         auxdata = timers_iter->auxdata;
      }

      // Return this vector of pointers to the free list
      stacklist_free(&sl_heap, timers_start);

      spin_unlock(&kerninst_cswitch_lock);
   }
}

void on_switch_out(void)
{
   int grabbed_freelistentry_yet = 0;

   void **iter, **finish;

   struct saved_timer *vec_of_timers = NULL;
   struct saved_timer *vec_of_timers_iter = NULL;
   struct saved_timer *vec_of_timers_finish = NULL;

   struct task_struct *curr = current;
   unsigned curr_thr = (unsigned)curr->pid;
   unsigned curr_cpu = (unsigned)smp_processor_id();

   if(cswitch_stuff_enabled) {
      spin_lock(&kerninst_cswitch_lock);

      iter = (void**)av_vtimers.vtimer_addresses;
      finish = (void**)av_vtimers.finish;

      // Loop through all vtimers
      while (iter != finish) {
         struct vtimer *theTimer = (struct vtimer *)*iter++;
         struct vtimer *cpu_vt = getCPUTimer(theTimer, curr_cpu);
         // read start field to get the state
         if(cpu_vt->start_field.rcount != 0) { // timer started
            if (!grabbed_freelistentry_yet) {
               vec_of_timers = (struct saved_timer*)stacklist_get(&sl_heap);
               vec_of_timers_iter = vec_of_timers;
               vec_of_timers_finish = vec_of_timers_iter + max_num_vtimers;
               grabbed_freelistentry_yet = 1;
            }
            // Call stop_routine, which updates saved_timer.auxdata, 
            // vtimer.total_field, and vtimer.start_field
            generic_vstop_routine(cpu_vt, vec_of_timers_iter, 
                                  theTimer->counter_type);
            
            // Add to working list:
            (struct vtimer*)(vec_of_timers_iter->addr) = theTimer;
            vec_of_timers_iter++;
         }
      }
      if (grabbed_freelistentry_yet) { // was anything done?
         // What to set the key to? id of current thread
         unsigned theKey = curr_thr;
	 
         // Mark the end of the list
         (struct vtimer*)(vec_of_timers_iter->addr) = NULL;

         // Add to the hash table!
         hash_table_add(&ht_vtimers, 
                        vec_of_timers, // not vec_of_timers_iter!
                        theKey);
      }
      spin_unlock(&kerninst_cswitch_lock);
   }
}
