// countedVEvents.h
// simple metric for counted, virtualized events, where an "event" is something
// like "machine cycles" or "# of cache misses".
// countedVEvents is just a base class; derived classes are needed to create
// primitives.  This allows much code to be shared.
//
// Compare to class threadEvents, which is similar except (1) it isn't counted, and
// thus isn't recursion safe, and (2) has different units (thread x events instead
// of just events)

#ifndef _COUNTED_VEVENTS_H_
#define _COUNTED_VEVENTS_H_

#include "util/h/kdrvtypes.h"
#include "simpleMetric.h"
#include "kapi.h"

class countedVEvents : public simpleMetric {
 private:
    kapi_hwcounter_kind hwkind;
    kapi_hwcounter_expr rawCtrExpr;

    // Initialize the timer data in the kernel memory
    void initialize_timer(kptr_t kernelAddr,
			  unsigned elemsPerVector,
			  unsigned bytesPerStride,
			  kptr_t stopRoutineKernelAddr,
			  kptr_t restartRoutineKernelAddr) const;

   // prevent copying
   countedVEvents(const countedVEvents &);
   countedVEvents& operator=(const countedVEvents &);
   
 public:
   countedVEvents(unsigned iSimpleMetId, kapi_hwcounter_kind iKind) :
      simpleMetric(iSimpleMetId), hwkind(iKind), rawCtrExpr(iKind) {
   }
   virtual ~countedVEvents() {
   }

   bool canInstantiateFocus(const focus &theFocus) const;
   unsigned getNumValues() const { return 1; }

   const kapi_hwcounter_expr &getRawCtrExpr() const {
       return rawCtrExpr;
   }

   // This routine determines whether to include events that occured 
   // while we were switched out. Wall timers want to include them
   virtual bool includeEventsSwitchedOut() const {
       return false;
   }

   // What is does: emits code to stop a vtimer of this metric (if needed), and
   // also appends the 8 byte {vtimeraddr, auxdata [rcount]} entity to
   // vtimerIter iff stopping. vtimerStart is passed in to aid detection of
   // overflow.
   kapi_snippet generateMetricSpecificVStop(
       const kapi_snippet &vtimer_addr_base,
       const kapi_snippet &vtimer_start,
       const kapi_snippet &vtimer_iter) const;
   // The opposite of the above routine; they work hand-in-hand.
   kapi_snippet generateMetricSpecificVRestart(
       const kapi_snippet &vtimer_addr_base,
       const kapi_snippet &vtimer_auxdata) const;

   void asyncInstantiate(simpleMetricFocus &,
                         cmfInstantiatePostSmfInstantiation &) const;

   void cleanUpAfterUnsplice(const simpleMetricFocus &smf) const;
      // Typically when a simpleMetricFocus is unspliced, all that's needed is to free
      // the associated data & code snippets.  But we have to do much more:
      // (1) remove vtimer from all_vtimers[] (so cswitch handlers won't try to stop/
      // restart the vtimer any more).
      // (2) decrement a refcount for how-to-stop and how-to-restart routines for
      // this simpleMetric; if now 0, then un-download them from kernel.  (Don't worry,
      // we can be sure they're not in-progress, since they're not interruptable except
      // by trap handlers or other stuff above LOCK_LEVEL.)

   // Get the state of the counters as if the metric has been enabled
   kapi_hwcounter_set queryPostInstrumentationPcr(
       const kapi_hwcounter_set &oldPcrValue, const focus &) const {
       
       kapi_hwcounter_set result = oldPcrValue;
       result.insert(hwkind);

       return result;
   }
   
   // Check if the metric can be enabled without invalidating another one
   bool changeInPcrWouldInvalidateSMF(
       const kapi_hwcounter_set &pendingNewPcrValue, const focus &) const {

       return pendingNewPcrValue.conflictsWith(hwkind);
   }
};

// ********************************************************************************

class countedVCycles : public countedVEvents {
 public:
   countedVCycles(unsigned iSimpleMetricId) :
       countedVEvents(iSimpleMetricId, HWC_TICKS) {
   }
};

class countedWallCycles : public countedVCycles {
 public:
    countedWallCycles(unsigned iSimpleMetricId) : 
	countedVCycles(iSimpleMetricId) {
    }
    bool includeEventsSwitchedOut() const {
	return true;
    }
};

#ifdef sparc_sun_solaris2_7
// ********************************************************************************

class countedDCacheVReads : public countedVEvents {
 public:
   countedDCacheVReads(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_DCACHE_VREADS) {
   }
};

// ********************************************************************************

class countedDCacheVWrites : public countedVEvents {
 public:
    countedDCacheVWrites(unsigned iSimpleMetricId) :
	countedVEvents(iSimpleMetricId, HWC_DCACHE_VWRITES) {
    }
};

// ********************************************************************************

class countedDCacheVReadHits : public countedVEvents {
 public:
   countedDCacheVReadHits(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_DCACHE_VREAD_HITS) {
   }
};

// ********************************************************************************

class countedDCacheVWriteHits : public countedVEvents {
 public:
   countedDCacheVWriteHits(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_DCACHE_VWRITE_HITS) {
   }
};

// ********************************************************************************

class countedICacheVRefs : public countedVEvents {
 public:
   countedICacheVRefs(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_ICACHE_VREFS) {
   }
};

// ********************************************************************************

class countedICacheVHits : public countedVEvents {
 public:
   countedICacheVHits(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_ICACHE_VHITS) {
   }
};

// ********************************************************************************

class countedICacheVStallCycles : public countedVEvents {
 public:
   countedICacheVStallCycles(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_ICACHE_STALL_CYCLES) {
   }
};

class countedBranchMispredVStallCycles : public countedVEvents {
 public:
   countedBranchMispredVStallCycles(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_BRANCH_MISPRED_VSTALL_CYCLES) {
   }
};


// ********************************************************************************

class countedECacheVRefs : public countedVEvents {
 public:
   countedECacheVRefs(unsigned iSimpleMetricId) :
       countedVEvents(iSimpleMetricId, HWC_ECACHE_VREFS) {
   }
};

// ********************************************************************************

class countedECacheVHits : public countedVEvents {
 public:
   countedECacheVHits(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_ECACHE_VHITS) {
   }
};

// ********************************************************************************

class countedECacheVReadHits : public countedVEvents {
 public:
   countedECacheVReadHits(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_ECACHE_VREAD_HITS) {
   }
};

// ********************************************************************************

class countedVInsnsPic0 : public countedVEvents {
   // #insns is a little unusual, in that it can use %pic0 event 0x1 **or**
   // %pic1 event 0x1.  This simpleMetric is for %pic0; another will be for %pic1
 public:
   countedVInsnsPic0(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_VINSNS) {
   }
};

// ********************************************************************************

class countedVInsnsPic1 : public countedVEvents {
   // #insns is a little unusual, in that it can use %pic1 event 0x1 **or**
   // %pic1 event 0x1.  This simpleMetric is for %pic1; above one is for %pic0

 public:
   countedVInsnsPic1(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_VINSNS) {
   }
};
#elif defined(i386_unknown_linux2_4)

class countedVInsnsPic0 : public countedVEvents {
   // #insns is a little unusual, in that it can use %pic0 event 0x1 **or**
   // %pic1 event 0x1.  This simpleMetric is for %pic0; another will be for %pic1
 public:
   countedVInsnsPic0(unsigned iSimpleMetricId) :
      countedVEvents(iSimpleMetricId, HWC_INSTR_ISSUED) {
   }
};

#endif // arch_os


#endif
