#include <string.h>
#include "metricVirtualizer.h"
#include "all_vtimers.h"
#include "kapi.h"

extern kapi_manager kmgr;

kptr_t metricVirtualizer::all_vtimers_create()
{
#ifdef sparc_sun_solaris2_7
    kptr_t rv = kmgr.malloc(sizeof_allvtimers);
    assert(rv != 0);

    // Prepare the structure locally
    char *buffer = new char[sizeof_allvtimers];
    assert(buffer != 0);
    memset(buffer, 0, sizeof_allvtimers);
    kptr_t finish = rv; // pointer to the first available entry
    kmgr.to_kernel_byte_order(&finish, sizeof(kptr_t));
    memcpy(&buffer[all_vtimers_offset_to_finish], &finish, sizeof(kptr_t));

    if (kmgr.memcopy(buffer, rv, sizeof_allvtimers) < 0) {
	assert(false);
    }
    delete[] buffer;

    return rv;
#else
    return kmgr.getAllVTimersAddr();
#endif
}

void metricVirtualizer::all_vtimers_delete(kptr_t all_vtimers_addr)
{
#ifdef sparc_sun_solaris2_7
    kmgr.free(all_vtimers_addr);
#endif
}

void metricVirtualizer::add_timer_to_all_vtimers(kptr_t all_vtimers_addr, 
						 kptr_t this_vtimer_addr)
{
#ifdef sparc_sun_solaris2_7
    // Read-in the pointer to the first available spot
    kptr_t finish;
    kmgr.memcopy(all_vtimers_addr + all_vtimers_offset_to_finish,
		 &finish, sizeof(kptr_t));
    kmgr.to_client_byte_order(&finish, sizeof(kptr_t));
    assert(finish >= all_vtimers_addr);
    // assert that we have enough space there
    // notice that the last entry in the list should always be 0 -> -sizeof()
    assert(finish < all_vtimers_addr + all_vtimers_offset_to_finish -
	   sizeof(kptr_t));

    // Write updated finish and this_vtimer_addr to the kernel space
    kptr_t new_finish = finish + sizeof(kptr_t);
    kmgr.to_kernel_byte_order(&new_finish, sizeof(kptr_t));
    kmgr.to_kernel_byte_order(&this_vtimer_addr, sizeof(kptr_t));

    kmgr.memcopy(&new_finish, all_vtimers_addr + all_vtimers_offset_to_finish,
		 sizeof(kptr_t));
    kmgr.memcopy(&this_vtimer_addr, finish, sizeof(kptr_t));
 
    // Assert that the list is still zero-terminated
    kptr_t zero;
    kmgr.memcopy(finish + sizeof(kptr_t), &zero, sizeof(kptr_t));
    assert(zero == 0);
#else
    kmgr.addTimerToAllVTimers(this_vtimer_addr);
#endif
}

void metricVirtualizer::remove_timer_from_all_vtimers(kptr_t all_vtimers_addr,
						      kptr_t this_vtimer_addr)
{
#ifdef sparc_sun_solaris2_7
    // Read-in all_vtimers
    kptr_t *buffer = new kptr_t[max_num_vtimers+1]; // +1 for finish

    assert(sizeof_allvtimers == (max_num_vtimers+1) * sizeof(kptr_t));
    if (kmgr.memcopy(all_vtimers_addr, buffer, sizeof_allvtimers) < 0) {
	assert(false);
    }

    int last_used = -1;
    for (; last_used+1 < max_num_vtimers && buffer[last_used+1] != 0; 
	 last_used++);
    assert(last_used >= 0);

    kptr_t finish = buffer[max_num_vtimers];
    kmgr.to_client_byte_order(&finish, sizeof(kptr_t));
    finish -= sizeof(kptr_t);
    assert(finish == all_vtimers_addr + last_used * sizeof(kptr_t));
    kmgr.to_kernel_byte_order(&finish, sizeof(kptr_t));

    // Convert it once, so we can compare raw data in the buffer to it
    kmgr.to_kernel_byte_order(&this_vtimer_addr, sizeof(kptr_t));
    
    // Find the timer to delete, write the last one in its place,
    // write 0 in the last place, update the finish pointer
    for (unsigned i = 0; i < max_num_vtimers && buffer[i] != 0; i++) {
	if (buffer[i] == this_vtimer_addr) {
	    kmgr.memcopy(&buffer[last_used], all_vtimers_addr + 
			 i * sizeof(kptr_t), sizeof(kptr_t));
	    kptr_t zero = 0;
	    kmgr.memcopy(&zero, all_vtimers_addr + 
			 last_used * sizeof(kptr_t), sizeof(kptr_t));
	    kmgr.memcopy(&finish, all_vtimers_addr + 
			 all_vtimers_offset_to_finish, sizeof(kptr_t));
	    delete[] buffer;

	    return;
	}
    }
    assert(false && "Timer not found");
#else
    kmgr.removeTimerFromAllVTimers(this_vtimer_addr);
#endif
}

void metricVirtualizer::instrument_switch_points(
    kapi_point_location type,
    const kapi_vtimer_cswitch_expr &cswitch_expr,
    kapi_vector<int> *callReqids, int *pCswitchReqid)
{
    kapi_vector<kapi_point> cswitchPoints;

    if (kmgr.findSpecialPoints(type, &cswitchPoints) < 0) {
	assert(false);
    }

#ifdef sparc_sun_solaris2_7
    int cswitch_reqid = kmgr.uploadSnippet(cswitch_expr, cswitchPoints);
    if (cswitch_reqid < 0) {
	assert(false);
    }
    *pCswitchReqid = cswitch_reqid;

    kptr_t cswitch_addr = kmgr.getUploadedAddress(cswitch_reqid);
#else
    kapi_module kdriver_mod;
    if(kmgr.findModule("kerninst", &kdriver_mod) < 0) {
	assert(!"could not locate kerninst module\n");
    }
    kapi_function kfunc;
    if(type == kapi_cswitch_in) {
       if (kdriver_mod.findFunction("on_switch_in", &kfunc) < 0) {
          assert(!"could not locate on_switch_in() function\n");
       }
    }
    else {
       if (kdriver_mod.findFunction("on_switch_out", &kfunc) < 0) {
          assert(!"could not locate on_switch_out() function\n");
       }
    }
    kptr_t cswitch_addr = kfunc.getEntryAddr();
#endif

    kapi_vector<kapi_snippet> empty;
    kapi_call_expr my_call(cswitch_addr, empty);

    kapi_vector<kapi_point>::const_iterator ipoint = cswitchPoints.begin();
    for (; ipoint != cswitchPoints.end(); ipoint++) {
	int id = kmgr.insertSnippet(my_call, *ipoint);
	if (id < 0) {
	    assert(false);
	}
	callReqids->push_back(id);
    }
}

void metricVirtualizer::deinstrument_switch_points(
    const kapi_vector<int> &call_reqid, int cswitch_reqid)
{
    kapi_vector<int>::const_iterator icall = call_reqid.begin();
    for (; icall != call_reqid.end(); icall++) {
	if (kmgr.removeSnippet(*icall) < 0) {
	    assert(false);
	}
    }
    
#ifdef sparc_sun_solaris2_7
    if (kmgr.removeUploadedSnippet(cswitch_reqid) < 0) {
	assert(false);
    }
#endif
}

metricVirtualizer::~metricVirtualizer() 
{
    assert(numReferences == 0);
    assert(allVTimersAddr == 0);
}

metricVirtualizer::metricVirtualizer() : numReferences(0), allVTimersAddr(0), 
    theSimpleMetricSpecificVRoutines(simpleMetricPtrHash)
{
}

void metricVirtualizer::referenceVirtualizationCode() 
{
    // If reference count was zero, installs a lot of cswitch-related stuff
    // into the kernel (cswitchout, cswitchin, blank all_vtimers[], blank
    // hash table, stacklist heap), and splices it in.

    if (numReferences++ == 0) {
	// Install all_vtimers: a vector to list all available vtimers
	assert(allVTimersAddr == 0);
	allVTimersAddr = all_vtimers_create();
	assert(allVTimersAddr != 0);
	
	// Expressions for the switch-in-and-out code
	cswitchin_expr = new kapi_vtimer_cswitch_expr(kapi_cswitch_in, 
						      allVTimersAddr);
	cswitchout_expr = new kapi_vtimer_cswitch_expr(kapi_cswitch_out,
						       allVTimersAddr);

	// Upload the code and splice it into the kernel
	instrument_switch_points(kapi_cswitch_in, *cswitchin_expr,
				 &callInReqid, &cswitchInReqid);

	instrument_switch_points(kapi_cswitch_out, *cswitchout_expr,
				 &callOutReqid, &cswitchOutReqid);
    }
}

void metricVirtualizer::dereferenceVirtualizationCode()
{
    assert(numReferences > 0);
    numReferences--;
    if (numReferences == 0) {
	deinstrument_switch_points(callOutReqid, cswitchOutReqid);
	deinstrument_switch_points(callInReqid, cswitchInReqid);

	all_vtimers_delete(allVTimersAddr);
	allVTimersAddr = 0;
	delete cswitchin_expr;
	delete cswitchout_expr;
	cswitchin_expr = cswitchout_expr = 0;
	callInReqid.clear();
	callOutReqid.clear();
    }
}

std::pair<kptr_t, kptr_t>
metricVirtualizer::referenceVRoutinesForMetric(
    const countedVEvents &theSimpleMetric) 
{
    // installs if needed; otherwise just bumps up refcount
    if (!theSimpleMetricSpecificVRoutines.defines(&theSimpleMetric)) {
	// We need to install the routines
	assert(theSimpleMetric.getNumValues() == 1);

#ifdef sparc_sun_solaris2_7
	kapi_snippet vrestart = theSimpleMetric.generateMetricSpecificVRestart(
	    cswitchin_expr->getVRestartAddrExpr(),
	    cswitchin_expr->getAuxdataExpr());
	kapi_snippet vstop = theSimpleMetric.generateMetricSpecificVStop(
	    cswitchout_expr->getVStopAddrExpr(),
	    cswitchout_expr->getStartExpr(),
	    cswitchout_expr->getIterExpr());

	kapi_vector<kapi_point> dummy_vrestart_point;
	dummy_vrestart_point.push_back(kapi_point(0, kapi_vrestart_routine));
	kapi_vector<kapi_point> dummy_vstop_point;
	dummy_vstop_point.push_back(kapi_point(0, kapi_vstop_routine));
	
	int vrestart_id, vstop_id;
	if ((vrestart_id = kmgr.uploadSnippet(vrestart,
					      dummy_vrestart_point)) < 0 ||
	    (vstop_id = kmgr.uploadSnippet(vstop,
					   dummy_vstop_point)) < 0) {
	    assert(false);
	}
	kptr_t vrestart_addr = kmgr.getUploadedAddress(vrestart_id);
	kptr_t vstop_addr = kmgr.getUploadedAddress(vstop_id);
#else
        int vrestart_id = 0;
        int vstop_id = 0;

        kapi_module kmod;
        if(kmgr.findModule("kerninst", &kmod) < 0) {
           assert(!"could not locate kerninst module\n");
        }
        kapi_function vrestart_func, vstop_func;
        if (kmod.findFunction("generic_vrestart_routine", &vrestart_func) < 0) {
           assert(!"could not locate generic_vrestart_routine() function\n");
        }
        if (kmod.findFunction("generic_vstop_routine", &vstop_func) < 0) {
           assert(!"could not locate generic_vstop_routine() function\n");
        }
        kptr_t vrestart_addr = vrestart_func.getEntryAddr();
        kptr_t vstop_addr = vstop_func.getEntryAddr();
#endif // ifdef sparc_sun_solaris2_7

	simpleMetricSpecificVRoutines theVRoutines;
	theVRoutines.refCount = 0;
	theVRoutines.stopIfNeededReqId = vstop_id;
	theVRoutines.stopIfNeededKernelAddr = vstop_addr;
	theVRoutines.restartReqId = vrestart_id;
	theVRoutines.restartKernelAddr = vrestart_addr;
	theSimpleMetricSpecificVRoutines.set(&theSimpleMetric, theVRoutines);
    }

    simpleMetricSpecificVRoutines &theVRoutines =
	theSimpleMetricSpecificVRoutines.get(&theSimpleMetric);
    ++theVRoutines.refCount;

    return std::make_pair(theVRoutines.stopIfNeededKernelAddr,
                          theVRoutines.restartKernelAddr);
}

void metricVirtualizer::dereferenceVRoutinesForMetric(
    const countedVEvents &theSimpleMetric) 
{
    simpleMetricSpecificVRoutines &theRoutines = 
	theSimpleMetricSpecificVRoutines.get(&theSimpleMetric);

    assert(theRoutines.refCount);
    if (--theRoutines.refCount == 0) {
#ifdef sparc_sun_solaris2_7
	if (kmgr.removeUploadedSnippet(theRoutines.restartReqId) < 0) {
	    assert(false);
	}
	if (kmgr.removeUploadedSnippet(theRoutines.stopIfNeededReqId) < 0) {
	    assert(false);
	}
#endif
	theSimpleMetricSpecificVRoutines.undef(&theSimpleMetric);
    }
}

void metricVirtualizer::virtualizeInitializedTimer(kptr_t vtimerAddrInKernel) 
{
    // invariants: vtimer exists and has had its fields fully
    // initialized, including the how-to-stop and how-to-restart fields (for
    // the benefit of cswitch handler routines).  cswitch stuff (hash table,
    // all_vtimers[], cswitchout/in, etc.)  have been spliced into the
    // kernel. We can now add vtimerAddrInKernel to all_vtimers[]

    assert(allVTimersAddr != 0);
    add_timer_to_all_vtimers(allVTimersAddr, vtimerAddrInKernel);
}

void metricVirtualizer::unVirtualizeVTimer(kptr_t vtimerAddrInKernel) 
{
    // Remove vtimerAddrInKernel from all_vtimers[]
    
    assert(allVTimersAddr != 0);
    remove_timer_from_all_vtimers(allVTimersAddr, vtimerAddrInKernel);
}
